package mv3302;

import java.awt.geom.Point2D;
import static java.lang.Double.NaN;
import simkit.Schedule;
import simkit.SimEntityBase;

/**
 *
 * @author ahbuss
 */
public class SimplestMover extends SimEntityBase {

    public static final Point2D NaP = new Point2D.Double(NaN, NaN);

    private Point2D initialLocation;

    private double maxSpeed;

    protected Point2D lastStopLocation;

    protected Point2D velocity;

    protected Point2D destination;

    protected double startMoveTime;

    public SimplestMover() {
        lastStopLocation = (Point2D) NaP.clone();
        velocity = (Point2D) NaP.clone();
        destination = (Point2D) NaP.clone();
    }

    public SimplestMover(Point2D initialLocation, double maxSpeed) {
        this();
        setInitialLocation(initialLocation);
        setMaxSpeed(maxSpeed);
    }

    @Override
    public void reset() {
        super.reset();
        lastStopLocation.setLocation(initialLocation);
        velocity.setLocation(0.0, 0.0);
        destination.setLocation(NaP);
        startMoveTime = 0.0;
    }

    public void doRun() {
        firePropertyChange("lastStopLocation", getLastStopLocation());
        firePropertyChange("velocity", getVelocity());
        firePropertyChange("destination", getDestination());
        firePropertyChange("startMoveTime", getStartMoveTime());
    }

    public void doMoveTo(Point2D newDestination) {
        Point2D oldDestination = getDestination();
        this.destination.setLocation(newDestination);
        firePropertyChange("destination", oldDestination, getDestination());

        if (lastStopLocation.distance(destination) > 0.0) {
            waitDelay("StartMove", 0.0);
        }
    }

    public void doStartMove() {
        double oldStartMoveTime = getStartMoveTime();
        startMoveTime = Schedule.getSimTime();
        firePropertyChange("startMoveTime", oldStartMoveTime, getStartMoveTime());

        double distance = lastStopLocation.distance(destination);
        double a = getMaxSpeed() / distance;

        Point2D oldVelocity = getVelocity();
        velocity.setLocation(destination.getX() - lastStopLocation.getX(), destination.getY() - lastStopLocation.getY());
        velocity.setLocation(a * velocity.getX(), a * velocity.getY());
        firePropertyChange("velocity", oldVelocity, getVelocity());

        waitDelay("EndMove", 1.0 / a);
    }

    public void doEndMove() {
        double oldStartMoveTime = getStartMoveTime();
        startMoveTime = Schedule.getSimTime();
        firePropertyChange("startMoveTime", oldStartMoveTime, getStartMoveTime());

        Point2D oldLastStopLocation = getLastStopLocation();
        lastStopLocation.setLocation(destination);
        firePropertyChange("lastStopLocation", oldLastStopLocation, getLastStopLocation());

        Point2D oldVelocity = getVelocity();
        velocity.setLocation(0.0, 0.0);
        firePropertyChange("velocity", oldVelocity, getVelocity());

    }

    /**
     * @return the initialLocation
     */
    public Point2D getInitialLocation() {
        return (Point2D) initialLocation.clone();
    }

    /**
     * @param initialLocation the initialLocation to set
     */
    public void setInitialLocation(Point2D initialLocation) {
        this.initialLocation = (Point2D) initialLocation.clone();
    }

    /**
     * @return the maxSpeed
     */
    public double getMaxSpeed() {
        return maxSpeed;
    }

    /**
     * @param maxSpeed the maxSpeed to set
     */
    public void setMaxSpeed(double maxSpeed) {
        if (maxSpeed < 0.0) {
            throw new IllegalArgumentException("maxSpeed must be >= 0.0: " + maxSpeed);
        }
        this.maxSpeed = maxSpeed;
    }

    /**
     * @return the lastStopLocation
     */
    public Point2D getLastStopLocation() {
        return (Point2D) lastStopLocation.clone();
    }

    /**
     * @return the velocity
     */
    public Point2D getVelocity() {
        return (Point2D) velocity.clone();
    }

    /**
     * @return the destination
     */
    public Point2D getDestination() {
        return (Point2D) destination.clone();
    }

    /**
     * @return the startMoveTime
     */
    public double getStartMoveTime() {
        return startMoveTime;
    }

    public Point2D getCurrentLocation() {
        double elapsedTime = Schedule.getSimTime() - getStartMoveTime();
        return new Point2D.Double(lastStopLocation.getX() + elapsedTime * velocity.getX() , 
            lastStopLocation.getY() + elapsedTime * velocity.getY());
    }
    
    @Override
    public String toString() {
        return String.format("%s %s", getName(), getCurrentLocation());
    }
    
    public String paramString() {
        return super.toString();
    }
}
